<?php
/* Copyright start
  Copyright (C) 2008 - 2020 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace Cybersponse\CoreBundle\Controller;

use DOMDocument;
use Doctrine\ORM\EntityManager;
use Cybersponse\UtilsBundle\ARCUtils;
use Dunglas\ApiBundle\JsonLd\Response;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\HttpFoundation\Request;
use Cybersponse\TenantBundle\Entity\Prod\Widget;
use Symfony\Component\HttpFoundation\File\UploadedFile;
use Symfony\Component\HttpKernel\Exception\HttpException;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Cybersponse\CoreBundle\Controller\AuthorizationResourceController;
use Cybersponse\AuthorizationBundle\Permissions\PermissionsDictionaryBuilder;

class WidgetController extends AuthorizationResourceController
{
    const TMP_PATH = '/tmp/widget';

    /** @var FileSystem */
    protected $fileSystem;

    
    public function installWidgetAction(Request $request)
    {
        $requestBody = ARCUtils::extractRequestBody($request, null);
        $resource = $this->getResource($request);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_CREATE);

        if (!$requestBody->name) {
            return new Response('Missing widget name', Response::HTTP_BAD_REQUEST);
        }

        if (!$requestBody->version) {
            return new Response('Missing widget version', Response::HTTP_BAD_REQUEST);
        }
        $client = new \GuzzleHttp\Client();
        try {
            $widgetName = $requestBody->name."-".$requestBody->version;
            $widgetRepoLink = sprintf('%s/widgets/%s/%s.tgz', getenv("REPOSERVER"), $widgetName, $widgetName);
            $responseObject = $client->request('GET', $widgetRepoLink, ['verify' => false]);
            $fileData = $responseObject->getBody()->getContents();
        } catch (\Exception $e) {
            throw new HttpException(Response::HTTP_NOT_FOUND, $e->getMessage(), $e);
        }

        $fileBName = $requestBody->name.'-'.$requestBody->version;
        $tgzFileName = sprintf('%s.tgz', $fileBName);

        $this->unzipWidgetFile($fileData, $tgzFileName, $fileBName);

        return $this->createWidget($request, $fileBName, $operation="install");
    }

    public function uploadWidgetAction(Request $request)
    {
        $resource = $this->getResource($request);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_CREATE);

        if (!$request->files->has('file')) {
            throw new BadRequestHttpException('No file specified.');
        }

        /** @var UploadedFile $upload */
        $upload = $request->files->get('file');
        
        $fileName = $upload->getClientOriginalName();
        $fileExt = $upload->getClientOriginalExtension();
        $fileBName = explode(".".$fileExt, $fileName)[0];

        $openFile = $upload->openFile();
        $readFile = $openFile->fread($openFile->getSize());
        
        $this->unzipWidgetFile($readFile, $fileName, $fileBName);
        
        return $this->createWidget($request, $fileBName, $operation = "upload");
    }

    public function revertWidgetAction(Request $request, $id)
    {
        $resource = $this->getResource($request);
        $this->denyAccessUnlessGranted($resource, PermissionsDictionaryBuilder::PERM_CRUD_UPDATE);
        /** @var EntityManager $entityManager */
       
        $entityManager = $this->container->get('doctrine')->getManagerForClass(Widget::class);
        $widgetRepository = $entityManager->getRepository(Widget::class);
        $widgetsFileSystem = new Filesystem();
        /** @var Widget $widget */
        $widget = $widgetRepository->find($id);
        if ($widget) {
            $widgetsPath = sprintf("/opt/cyops-ui/widgets/%s", $widget->getName());
  
            if ($widgetsFileSystem->exists($widgetsPath)) {
                try {
                    $widgetsHtml = sprintf('%s/html/view.html', $widgetsPath, $widget->getName());
                    $widgetsController = sprintf('%s/js/view.controller.js', $widgetsPath, $widget->getName());
                    $widgetsEditHtml = sprintf('%s/html/edit.html', $widgetsPath, ucfirst($widget->getName()));
                    $widgetsEditController = sprintf('%s/js/edit.controller.js', $widgetsPath, ucfirst($widget->getName()));
                    $data = [];
                    $data["viewHtml"] = file_get_contents($widgetsHtml);
                    $data["viewController"] = file_get_contents($widgetsController);
                    $data["editHtml"] =  file_get_contents($widgetsEditHtml);
                    $data["editController"] = file_get_contents($widgetsEditController);
                    $request->initialize(
                        $request->query->all(),
                        $request->request->all(),
                        $request->attributes->all(),
                        $request->cookies->all(),
                        $request->files->all(),
                        $request->server->all(),
                        json_encode($data)
                    );
                    return parent::putAction($request, $id);
                } catch (\Exception $e) {
                    throw new HttpException(Response::HTTP_BAD_REQUEST, $e->getMessage(), $e);
                }
            }
        }
        return new Response($this->container->getParameter("resource_not_found"), Response::HTTP_NOT_FOUND);
    }

    protected function createWidget($request, $fileBName, $operation = "upload")
    {
        $requestBody = $this->buildWidgetPayload($fileBName);
        if (isset($requestBody['error']) && isset($requestBody['error_msg'])) {
            return new Response($requestBody['error_msg'], Response::HTTP_BAD_REQUEST);
        }
        if ($operation == 'install') {
            $requestBody['draft'] = false;
            $requestBody['installed'] = true;
        }elseif ($operation == 'upload') {
            $requestBody['draft'] = true;
            $requestBody['installed'] = false;
            $timestamp = (new \DateTime('NOW', new \DateTimeZone('UTC')))->getTimestamp();
            $requestBody['published_date']=$timestamp;
            $requestBody['metadata']['certified'] = "No";
            $requestBody['metadata']['publisher'] = "Community";
        }
        $request->initialize(
            $request->query->all(),
            $request->request->all(),
            $request->attributes->all(),
            $request->cookies->all(),
            $request->files->all(),
            $request->server->all(),
            json_encode($requestBody)
        );
        $this->removeTmpFiles($fileBName);
         /** @var EntityManager $entityManager */
        $entityManager = $this->container->get('doctrine')->getManagerForClass(Widget::class);
        $widgetRepository = $entityManager->getRepository(Widget::class);
        $widget = $widgetRepository->findOneBy(['name' => $requestBody['name']]);
        if($widget)
        {
            return parent::putAction($request,$widget->getUuid());
        }
        return parent::cpostAction($request);
    }

    protected function unzipWidgetFile($tgzFileData, $tgzFileName, $fileName="")
    {
        try {
            $this->fileSystem = new Filesystem();
            if ($this->fileSystem->exists(self::TMP_PATH."/".$fileName)) {
                $this->removeTmpFiles($fileName);
            }
            $this->fileSystem->dumpFile(sprintf('%s/%s', self::TMP_PATH, $tgzFileName), $tgzFileData);
            // decompress from gz
            $phar = new \PharData(sprintf('%s/%s', self::TMP_PATH, $tgzFileName));
            $phar->decompress();
            $phar->extractTo(self::TMP_PATH);
        } catch (\Exception $e) {
            $this->removeTmpFiles($fileName);
            throw new HttpException(Response::HTTP_BAD_REQUEST, $e->getMessage(), $e);
        }
    }

    protected function buildWidgetPayload($name)
    {
        $data = array();
        try {
            $fPath = sprintf('%s/%s/info.json', self::TMP_PATH, $name);
            $fOpen = fopen($fPath, "r");
            $meta = fread($fOpen, filesize($fPath));
            foreach (json_decode($meta, true) as $k => $v) {
                $data[$k] = $v;
            }
        } catch (\Exception $e) {
            throw new HttpException(Response::HTTP_BAD_REQUEST, $e->getMessage(), $e);
        }
       
        $widgetFiles = array(
            'viewHtml' => $name.'/html/view.html',
            'viewController' => $name.'/js/view.controller.js',
            'editHtml' => $name.'/html/edit.html',
            'editController' => $name.'/js/edit.controller.js'
        );

        foreach ($widgetFiles as $key => $value) {
            try {
                $fPath = sprintf('%s/%s', self::TMP_PATH, $value);
                $fOpen = fopen($fPath, "r");

                if ($key == "info") {
                    $meta = fread($fOpen, filesize($fPath));
                    foreach (json_decode($meta) as $k => $v) {
                        $data[$k] = $v;
                    }
                } else {
                    $data[$key] = fread($fOpen, filesize($fPath));
                }
            } catch (\Exception $e) {
                $this->removeTmpFiles($name);
                throw new HttpException(Response::HTTP_BAD_REQUEST, $e->getMessage(), $e);
            }
        }

        $validCheck = $this->widgetValidation($data);
        if ($validCheck['error']) {
            $this->removeTmpFiles($name);
            return $validCheck;
        } else {
            return $data;
        }
    }

    protected function widgetValidation($data, $pathName=[])
    {
        try {
            $domHtml = new DOMDocument;
            $domHtml->loadHTML($data['viewHtml']);
        } catch (\Exception $e) {
            return array('error'=>true, 'error_msg'=>$this->container->getParameter('html_syntax_error'));
        }

        try {
            $domEditHtml = new DOMDocument;
            $domEditHtml->loadHTML($data['editHtml']);
        } catch (\Exception $e) {
            return array('error'=>true, 'error_msg'=>$this->container->getParameter('edithtml_syntax_error'));
        }
    }

    protected function removeTmpFiles($name)
    {
        $filePaths = [sprintf('%s/%s.tgz', self::TMP_PATH, $name),sprintf('%s/%s.tar', self::TMP_PATH, $name),sprintf('%s/%s', self::TMP_PATH, $name),sprintf('%s/._%s', self::TMP_PATH, $name)];
        foreach ($filePaths as $key => $value) {
            if ($this->fileSystem->exists($value)) {
                $this->fileSystem->remove($value);
            }
        }
    }

    public function deleteWidgetsAction(Request $request)
    {
        $requestResource = $this->getResource($request);
        $this->denyAccessUnlessGranted($requestResource, PermissionsDictionaryBuilder::PERM_CRUD_DELETE);
        $resource = ARCUtils::extractRequestBody($request, null);
        $partialUpdatedObject =[];
        $partialUpdatedObject["success"]=[];
        $partialUpdatedObject["failure"]=[];

        /** @var EntityManager $entityManager */
       
        $entityManager = $this->container->get('doctrine')->getManagerForClass(Widget::class);
        $widgetRepository = $entityManager->getRepository(Widget::class);
        $widgetsFileSystem = new Filesystem();
        foreach ($resource->ids as $id) {
            /** @var Widget $widget */
            $widget = $widgetRepository->find($id);
            if ($widget) {
                $widgetsPath = sprintf("/opt/cyops-ui/widgets/%s", $widget->getName());

                try {
                    parent::deleteAction($request, $id);
                    array_push($partialUpdatedObject["success"], $widget->getName());
                } catch (\Exception $e) {
                    $err[$widget->getName()] = $e->getMessage();
                    array_push($partialUpdatedObject["failure"], $err);
                    continue;
                } 
            }
        }
        if (count($partialUpdatedObject["failure"])>0) {
            return new Response($partialUpdatedObject, Response::HTTP_MULTI_STATUS);
        }
        return new Response($partialUpdatedObject, Response::HTTP_OK);
    }
}
